/*
 * GameConstants.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

#include "constants/NumericConstants.h"
#include "constants/Enumerations.h"
#include "constants/EntityIdentifiers.h"

VCMI_LIB_NAMESPACE_BEGIN

using TExpType = si64;
using TQuantity = si32;
using TRmgTemplateZoneId = int;

VCMI_LIB_NAMESPACE_END
