/*
 * MouseButton.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

enum class MouseButton
{
	LEFT   = 1,
	MIDDLE = 2,
	RIGHT  = 3,
	EXTRA1 = 4,
	EXTRA2 = 5
};
