/*
 * StdInc.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

#include "../../Global.h"

// This header should be treated as a pre compiled header file(PCH) in the compiler building settings.

// Here you can add specific libraries and macros which are specific to this project.

VCMI_LIB_USING_NAMESPACE
