/*
 * EOpenWindowMode.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

VCMI_LIB_NAMESPACE_BEGIN

enum class EOpenWindowMode : uint8_t
{
	EXCHANGE_WINDOW,
	RECRUITMENT_FIRST,
	RECRUITMENT_ALL,
	SHIPYARD_WINDOW,
	THIEVES_GUILD,
	UNIVERSITY_WINDOW,
	HILL_FORT_WINDOW,
	MARKET_WINDOW,
	PUZZLE_MAP,
	TAVERN_WINDOW
};

VCMI_LIB_NAMESPACE_END
