/*
 * EFonts.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

enum EFonts : int8_t
{
	FONT_BIG,
	FONT_CALLI,
	FONT_CREDITS,
	FONT_HIGH_SCORE,
	FONT_MEDIUM,
	FONT_SMALL,
	FONT_TIMES,
	FONT_TINY,
	FONT_VERD
};
