/*
 * MapObjects.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

// Helper header that includes all map objects, similar to old CObjectHandler.h
// Possible TODO - remove this header after CObjectHandler.cpp will be fully split into smaller files
#include "CObjectHandler.h"

#include "CArmedInstance.h"
#include "CGDwelling.h"
#include "CGHeroInstance.h"
#include "CGMarket.h"
#include "CGTownInstance.h"
#include "CGPandoraBox.h"
#include "CRewardableObject.h"
#include "MiscObjects.h"
#include "CQuest.h"
