/*
 * AUTHORS.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#pragma once

//VCMI PROJECT CODE CONTRIBUTORS:
const std::vector<std::vector<std::string>> contributors = {
//   Task          Name                    Aka                      E-Mail
   { "Idea",       "Michał Urbańczyk",     "Tow",                   "impono@gmail.com"             },
   { "Idea",       "Mateusz B.",           "Tow dragon",            "matcio1@gmail.com"            },

   { "Developing", "Andrea Palmate",       "afxgroup",              "andrea@amigasoft.net"         },
   { "Developing", "Alexander Shishkin",   "alexvins",              ""                             },
   { "Developing", "Rafal R.",             "ambtrip",               "ambtrip@wp.pl"                },
   { "Developing", "Andrii Danylchenko",   "",                      ""                             },
   { "Developing", "Benjamin Gentner",     "beegee",                ""                             },
   { "Developing", "Piotr Wójcik",         "Chocimier",             "chocimier@tlen.pl"            },
   { "Developing", "Dmitry Orlov",         "",                      "shubus.corporation@gmail.com" },
   { "Developing", "",                     "Dydzio",                "blood990@gmail.com"           },
   { "Developing", "Andrzej Żak",          "godric3",               ""                             },
   { "Developing", "Henning Koehler",      "henningkoehlernz",      "henning.koehler.nz@gmail.com" },
   { "Developing", "Ivan Savenko",         "",                      "saven.ivan@gmail.com"         },
   { "Developing", "",                     "kambala-decapitator",   "decapitator@ukr.net"          },
   { "Developing", "",                     "krs0",                  ""                             },
   { "Developing", "",                     "Laserlicht",            ""                             },
   { "Developing", "Alexey",               "Macron1Robot",          ""                             },
   { "Developing", "Michał Kalinowski",    "",                      "feniks_fire@o2.pl"            },
   { "Developing", "Vadim Glazunov",       "neweagle",              "neweagle@gmail.com"           },
   { "Developing", "Andrey Cherkas",       "nordsoft",              "nordsoft@yahoo.com"           },
   { "Developing", "Rickard Westerlund",   "Onion Knight",          "onionknigh@gmail.com"         },
   { "Developing", "Yifeng Sun",           "phoebus118",            "pkusunyifeng@gmail.com"       },
   { "Developing", "",                     "rilian-la-te",          ""                             },
   { "Developing", "",                     "SoundSSGood",           ""                             },
   { "Developing", "Stefan Pavlov",        "Ste",                   "mailste@gmail.com"            },
   { "Developing", "Arseniy Shestakov",    "SXX",                   "me@arseniyshestakov.com"      },
   { "Developing", "Lukasz Wychrystenko",  "tezeriusz",             "t0@czlug.icis.pcz.pl"         },
   { "Developing", "Trevor Standley",      "tstandley",             ""                             },
   { "Developing", "Vadim Markovtsev",     "",                      "gmarkhor@gmail.com"           },
   { "Developing", "Frank Zago",           "ubuntux",               ""                             },
   { "Developing", "",                     "vmarkovtsev",           ""                             },
   { "Developing", "Tom Zielinski",        "Warmonger",             "Warmonger@vp.pl"              },
   { "Developing", "Xiaomin Ding",         "",                      "dingding303@gmail.com"        },
   { "Developing", "Fenghuang Rumeng",     "kdmcser",               "zqtndfj@gmail.com"             },

   { "Testing",    "Ben Yan",              "by003",                 "benyan9110@gmail.com,"        },
   { "Testing",    "",                     "Misiokles",             ""                             },
   { "Testing",    "",                     "Povelitel",             ""                             },
};
