/*
* Goals.h, part of VCMI engine
*
* Authors: listed in file AUTHORS in main folder
*
* License: GNU General Public License v2.0 or later
* Full text of license available in license.txt file, in main folder
*
*/
#pragma once

#include "CGoal.h"
#include "Invalid.h"
#include "BuildBoat.h"
#include "BuildThis.h"
#include "BuyArmy.h"
#include "Trade.h"
#include "RecruitHero.h"
#include "DigAtTile.h"
#include "AdventureSpellCast.h"
