/*
 * CHeroClassTest.cpp, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#include "StdInc.h"

#include "../../lib/entities/hero/CHeroClass.h"

namespace test
{

using namespace ::testing;

class CHeroClassTest : public Test
{
public:
	MOCK_METHOD4(registarCb, void(int32_t, int32_t, const std::string &, const std::string &));

protected:
	std::shared_ptr<CHeroClass> subject;

	void SetUp() override
	{
		subject = std::make_shared<CHeroClass>();
	}
};

TEST_F(CHeroClassTest, RegistersIconsDoesNothing)
{
	auto cb = [this](auto && PH1, auto && PH2, auto && PH3, auto && PH4) 
	{
		registarCb(std::forward<decltype(PH1)>(PH1), std::forward<decltype(PH2)>(PH2), std::forward<decltype(PH3)>(PH3), std::forward<decltype(PH4)>(PH4));
	};
	subject->registerIcons(cb);
}

}
