/*
 * PossibleSpellcast.cpp, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */
#include "StdInc.h"

#include "PossibleSpellcast.h"

PossibleSpellcast::PossibleSpellcast()
	: spell(nullptr),
	dest(),
	value(0)
{
}

PossibleSpellcast::~PossibleSpellcast() = default;
