/*
* Goals.h, part of VCMI engine
*
* Authors: listed in file AUTHORS in main folder
*
* License: GNU General Public License v2.0 or later
* Full text of license available in license.txt file, in main folder
*
*/
#pragma once

#include "CGoal.h"
#include "Invalid.h"
#include "BuildBoat.h"
#include "Build.h"
#include "BuildThis.h"
#include "Conquer.h"
#include "GatherArmy.h"
#include "Win.h"
#include "VisitObj.h"
#include "VisitTile.h"
#include "VisitHero.h"
#include "Explore.h"
#include "BuyArmy.h"
#include "GatherTroops.h"
#include "Trade.h"
#include "CollectRes.h"
#include "RecruitHero.h"
#include "GetArtOfType.h"
#include "ClearWayTo.h"
#include "DigAtTile.h"
#include "FindObj.h"
#include "CompleteQuest.h"
#include "AdventureSpellCast.h"